------------ Grid and ACE stuff ------------


local GridDynamicLayout = AceLibrary("AceAddon-2.0"):new("AceConsole-2.0", "AceEvent-2.0")
local Aura = AceLibrary("SpecialEvents-Aura-2.0")
local BS = AceLibrary("Babble-Spell-2.2")
local L = AceLibrary("AceLocale-2.2"):new("GridDynamicLayout")

L:RegisterTranslations("enUS", function()
	return {
		["Dynamic Layout"] = true,
		["Options for GridDynamicLayout."] = true,
		["Player names list separated by commas, those will always appear in the tank group"]=true,
		["Player names list separated by commas, those will always appear in the heal group"]=true,
		["Player names list separated by commas, those will always appear in the ranged DPS group"]=true,
		["Player names list separated by commas, those will always appear in the melee DPS group"]=true,
		["Defaulted DPS warriors"]=true,
		["Player names list separated by commas, those will be defaulted as DPS warriors"]=true,
		["Forced Tanks"]=true,
		["Forced Heals"]=true,
		["Forced Melee DPS"]=true,
		["Forced Ranged DPS"]=true,
	}
end)

GridDynamicLayout = Grid:NewModule("GridDynamicLayout")
GridDynamicLayout.menuName = L["Dynamic Layout"]


------------ Locals ------------

local GROUPINDEX_TANK = 1
local GROUPINDEX_HEAL = 2
local GROUPINDEX_RDPS = 3
local GROUPINDEX_MDPS = 4
local dpsWarList = {
}

------------ GridDynamicLayoutACE Options -------------

GridDynamicLayout.options = {
	ForcedTanks = {
		type = "text",
			name = L["Forced Tanks"],
			desc = L["Player names list separated by commas, those will always appear in the tank group"],
			get = function() return GridDynamicLayout:generateOptionValue(GridDynamicLayout.db.profile.tankList) end,
			usage = L["Player names list separated by commas, those will always appear in the tank group"],
			set = function(v)
				GridDynamicLayout.db.profile.tankList = GridDynamicLayout:parseOptionValue(v)
				GridDynamicLayout:initializeLayout()
			end,
			order = 4,
	},
	ForcedHeals = {
		type = "text",
			name = L["Forced Heals"],
			desc = L["Player names list separated by commas, those will always appear in the heal group"],
			get = function() return GridDynamicLayout:generateOptionValue(GridDynamicLayout.db.profile.healList) end,
			usage = L["Player names list separated by commas, those will always appear in the heal group"],
			set = function(v)
				GridDynamicLayout.db.profile.healList = GridDynamicLayout:parseOptionValue(v)
				GridDynamicLayout:initializeLayout()
			end,
			order = 5,
	},
	ForcedMDPS = {
		type = "text",
			name = L["Forced Melee DPS"],
			desc = L["Player names list separated by commas, those will always appear in the melee DPS group"],
			get = function() return GridDynamicLayout:generateOptionValue(GridDynamicLayout.db.profile.MDPSList) end,
			usage = L["Player names list separated by commas, those will always appear in the melee DPS group"],
			set = function(v)
				GridDynamicLayout.db.profile.MDPSList = GridDynamicLayout:parseOptionValue(v)
				GridDynamicLayout:initializeLayout()
			end,
			order = 6,
	},
	ForcedRDPS = {
		type = "text",
			name = L["Forced Ranged DPS"],
			desc = L["Player names list separated by commas, those will always appear in the ranged DPS group"],
			get = function() return GridDynamicLayout:generateOptionValue(GridDynamicLayout.db.profile.RDPSList) end,
			usage = L["Player names list separated by commas, those will always appear in the ranged DPS group"],
			set = function(v)
				GridDynamicLayout.db.profile.RDPSList = GridDynamicLayout:parseOptionValue(v)
				GridDynamicLayout:initializeLayout()
			end,
			order = 7,
	},
	DefaultDPSWar = {
		type = "text",
			name = L["Defaulted DPS warriors"],
			desc = L["Player names list separated by commas, those will be defaulted as DPS warriors"],
			get = function() return GridDynamicLayout:generateOptionValue(GridDynamicLayout.db.profile.defaultDPSWarList) end,
			usage = L["Player names list separated by commas, those will be defaulted as DPS warriors"],
			set = function(v)
				GridDynamicLayout.db.profile.defaultDPSWarList = GridDynamicLayout:parseOptionValue(v)
				GridDynamicLayout:initializeLayout()
			end,
			order = 8,
	},
}

function GridDynamicLayout:parseOptionValue(s)
	local result = {}
	if string.find(s, ",") then
		for a in gmatch(s..",", "(.-),") do
			if a ~= "" then
				result[a] = true
			end
		end
	elseif s ~= "" then
		result[s] = true
	end
	return result
end

function GridDynamicLayout:generateOptionValue(list)
	local result = ""
	for key,value in pairs(list) do
		if value then
			result = result .. key .. ","
		end
	end
	return result
end

-- Hook into GridLayout options
GridLayout.options.args["DynamicLayout"] = {
	type = "group",
	name = L["Dynamic Layout"],
	desc = L["Options for GridDynamicLayout."],
	disabled = InCombatLockdown,
	args = GridDynamicLayout.options
}

------------ GridDynamicLayoutDefaultDB -------------

GridDynamicLayout.defaultDB = {
	debug=false,
	tankList = {},
	healList = {},
	MDPSList = {},
	RDPSList = {},
	defaultDPSWarList = {},
	columnSpacing = 2,
	unitsPerColumn = 5,
	maxColumns = 8,
}

------------ GridDynamicLayout ------------

local dynamicLayout

function GridDynamicLayout:OnInitialize()
	self.super.OnInitialize(self)
	dpsWarList = self.db.profile.defaultDPSWarList
end

function GridDynamicLayout:OnEnable()
	self:RegisterEvent("Grid_JoinedBattleground", "initializeLayout")
	self:RegisterEvent("Grid_JoinedRaid", "initializeLayout")
	self:RegisterEvent("SpecialEvents_UnitBuffLost", "checkBuff")
	self:RegisterEvent("SpecialEvents_UnitBuffGained", "checkBuff")
	self:RegisterEvent("Grid_UnitLeft", "initializeLayout")
	self:RegisterEvent("Grid_UnitJoin", "initializeLayout")
	self:RegisterEvent("CHAT_MSG_SPELL_PERIODIC_PARTY_BUFFS", "handleWarStanceChange")
end

function GridDynamicLayout:handleWarStanceChange(msg)
	local unitName, buff
	for unitName, buff in string.gmatch(msg, "(.+) gains (.+).") do
		
		if unitName and buff then
			if unitName=="You" then unitName = UnitName(player) end
			if buff == BS["Berserker Stance"] or buff == BS["Battle Stance"] then
				dpsWarList[unitName]=true
				self:initializeLayout()
			elseif buff == BS["Defensive Stance"] then
				dpsWarList[unitName]=false
				self:initializeLayout()
			end 
		end
	end
end

function GridDynamicLayout:addToGroup(grpNumber, unit)
	local name, realm = UnitName(unit)
	
	if realm ~= nil and realm ~= "" then
	name = name .. "-" .. realm
	end
	dynamicLayout[grpNumber].nameList = dynamicLayout[grpNumber].nameList .. name .. ","
end

function GridDynamicLayout:initializeLayout()
	local i
	
	self:Debug("Initializing Layout")
	
	for _,v in pairs(dynamicLayout) do
		v.nameList = ""
	end
	dynamicLayout.defaults.unitsPerColumn = self.db.profile.unitsPerColumn
	dynamicLayout.defaults.maxColumns = self.db.profile.maxColumns
	dynamicLayout.defaults.columnSpacing = self.db.profile.columnSpacing
	
	-- Scans the raid and place each unit in appropriate nameList
	for i=1,40 do
		if UnitExists("raid"..i) then
			self:addToGroup(self:getLayoutGroupIndex("raid"..i), "raid"..i)
		end
	end
	
	if GridLayout.db.profile.layout == "Dynamic Layout" then
		if not InCombatLockdown() then
			GridLayout:ReloadLayout()
		elseif not self.ExitingCombatScheduled then
			self:ScheduleLeaveCombatAction(GridDynamicLayout, "ExitingCombat")
			self.ExitingCombatScheduled = true
		end
	end
end



function GridDynamicLayout:ExitingCombat()
	self:ScheduleLeaveCombatAction(GridLayout, "ReloadLayout")
	self.ExitingCombatScheduled = false
	self:Debug("Performing ExitingCombat")
end

function GridDynamicLayout:getLayoutGroupIndex(unit)
	-- Add to tanks if "by name" setting used
	if self.db.profile.tankList[UnitName(unit)] then
		return GROUPINDEX_TANK
	-- Add to healers if "by name" setting used
	elseif self.db.profile.healList[UnitName(unit)] then
		return GROUPINDEX_HEAL
	-- Add to ranged DPS if "by name" setting used
	elseif self.db.profile.RDPSList[UnitName(unit)] then
		return GROUPINDEX_RDPS
	-- Add to melee DPS if "by name" setting used
	elseif self.db.profile.MDPSList[UnitName(unit)] then
		return GROUPINDEX_MDPS
	-- Use class / buff setting to assign role
	else
		local class
		_, class = UnitClass(unit)
		
		-- Add Shamans and Paladins to heal group
		if class=="SHAMAN" then
			return GROUPINDEX_HEAL
		-- Add Hunters, Mages and Warlocks to ranged DPS group
		elseif class=="HUNTER" or class=="MAGE" or class=="WARLOCK" then
			return GROUPINDEX_RDPS
		-- Add Rogues to melee DPS group
		elseif class=="ROGUE" then
			return GROUPINDEX_MDPS
		elseif class=="PALADIN" then
			-- Paladins to heal or tank if righteous fury --
		if Aura:UnitHasBuff(unit, BS["Righteous Fury"]) or Aura:UnitHasBuff(unit, BS["Improved Righteous Fury"]) then
				return GROUPINDEX_TANK
			else
				return GROUPINDEX_HEAL
			end
		elseif class=="WARRIOR" then
			-- Handling of warriors (cannot detect stance by normal means) --
			if dpsWarList[UnitName(unit)] then
				return GROUPINDEX_MDPS
			else
				return GROUPINDEX_TANK
			end
		elseif class=="PRIEST" then
			if Aura:UnitHasBuff(unit, BS["Shadowform"]) then
				return GROUPINDEX_RDPS
			else
				return GROUPINDEX_HEAL
			end
		elseif class=="DRUID" then
			if Aura:UnitHasBuff(unit, BS["Dire Bear Form"]) or Aura:UnitHasBuff(unit, BS["Bear Form"]) then
				return GROUPINDEX_TANK
			elseif Aura:UnitHasBuff(unit, BS["Cat Form"]) then
				return GROUPINDEX_MDPS
			elseif Aura:UnitHasBuff(unit, BS["Moonkin Form"]) then
				return GROUPINDEX_RDPS
			else
				return GROUPINDEX_HEAL
			end
		end
	end
end

function GridDynamicLayout:checkBuff(unit, name, count, icon, rank, duration)
	local class
	
	_, class = UnitClass(unit)
	
	if	(class=="PRIEST" and name==BS["Shadowform"]) or 
		(class=="DRUID" and (name==BS["Dire Bear Form"] or name==BS["Cat Form"] or name==BS["Moonkin Form"] or name==BS["Bear Form"])) or
		(class=="PALADIN" and (name==BS["Righteous Fury"] or name==BS["Improved Righteous Fury"]))
		then
			self:initializeLayout()
	end
end

dynamicLayout = {
	defaults = {
		unitsPerColumn = UNITS_PER_COLUMN,
		maxColumns = MAX_COLUMNS,
		columnAnchorPoint = "LEFT",
		sortMethod="INDEX",
	},
	[GROUPINDEX_TANK] = {					 -- tanks
		nameList = "",
	},
	[GROUPINDEX_HEAL] = {					 -- healers
		nameList = "",
	},
	[GROUPINDEX_RDPS] = {					 -- ranged DPS
		nameList = "",
	},
	[GROUPINDEX_MDPS] = {					 -- melee DPS
		nameList = "",
	},
}
GridLayout:AddLayout("Dynamic Layout", dynamicLayout)
